import maya.cmds as cmds
import math
import re
import random

##=====================================
# Useage: 
# import mesh_utils
# print mesh_utils.get_verts('pSphere1')
# Returns a list of lists ie. [[x,y,z],[x,y,z]....] 
##===================================== 
def get_verts(tnode, useLocalSpace=True):
	selected_verts = cmds.filterExpand(selectionMask=31)
	verts = []
	if selected_verts != None and len(selected_verts) > 0:
		for vertex in selected_verts:
			#print(vertex)
			vert_pos = cmds.pointPosition(vertex,local=useLocalSpace)
			verts.append(vert_pos)
		return verts
	# No selected vertices, therefore, we return all the mesh vertices
	shape = cmds.listRelatives(tnode, shapes=True)[0]
	num_verts = cmds.polyEvaluate(tnode, vertex=True)
	for n in range(num_verts):
		vert_str = shape + '.vtx[%d]' % n;
		vert_pos = cmds.pointPosition(vert_str, local=useLocalSpace) 
		verts.append(vert_pos)    
	return verts
	
##=====================================
# Same as get_verts() but returns a flat list of coords ie. 
#  [x0,y0,z0,x1,y1,z1....]
def get_coordinates(tnode, useLocalSpace=True):
	coords = []
	verts = get_verts(tnode, useLocalSpace)
	for vert in verts:
		coords.extend(vert)
	return coords

##=====================================
# Useage: 
# import mesh_utils
# print mesh_utils.get_normals('pSphere1')
# Returns a list of lists ie. [[x,y,z],[x,y,z]....] 
def get_normals(tnode):
	normals = []
	shape = cmds.listRelatives(tnode, shapes=True)[0]
	
	selected_verts = cmds.filterExpand(selectionMask=31)
	if selected_verts != None and len(selected_verts) > 0:
		for vert in selected_verts:
			coords = cmds.polyNormalPerVertex(vert, q=True, xyz=True)
			ave = get_average_normal(coords)
			normals.append(ave)
		return normals
	
	num_verts = cmds.polyEvaluate(tnode, vertex=True)
	for n in range(num_verts):
		vert_str = shape + '.vtx[%d]' % n;
		coords = cmds.polyNormalPerVertex(vert_str, q=True, xyz=True)
		ave = get_average_normal(coords)
		normals.append(ave)
	return normals

##=====================================
# used locally by get_normals()
##=====================================
def get_average_normal(coords):
	num_comps = len(coords)
	num_norms = num_comps/3
	if num_norms == 1:
		x,y,z = coords
		return [x,y,z]
	x = 0
	y = 0
	z = 0
	for n in range(0, num_comps - 3, 3):
		x += coords[n]
		y += coords[n + 1]
		z += coords[n + 2]
	return [x/num_norms, y/num_norms, z/num_norms]
	
##=====================================
def aim_y(vec):
	x,y,z = vec
	xy_length = math.sqrt(x * x + y * y)
	vec_length= math.sqrt(x * x + y * y + z * z)

	x_rot = math.acos(xy_length/vec_length)
	x_rot = x_rot if z > 0 else -x_rot
	x_rot = math.degrees(x_rot)
		
	# xy_length will be zero when vec is pointing
	# along the +z or -z axis
	if xy_length == 0:
		z_rot = math.radians(90) if x > 0 else math.radians(-90)
	else:
		z_rot = math.acos(y/xy_length)
	z_rot = -z_rot if x > 0 else z_rot
	z_rot = math.degrees(z_rot)
	return [x_rot, z_rot]
	##=====================================
# Thanks to Joseph Chen for supplying the code for the
# next function.
def get_faces_verts(tnode, useLocalSpace=True):
	faces = []
	shapes = cmds.listRelatives(tnode, shapes=True)
	if shapes == None:
		cmds.error('Cannot handle the group named "%s".' % tnode)
		return
		
	shape = shapes[0]
	num_faces = cmds.polyEvaluate(tnode, face=True)
	for n in range(num_faces):
		face_str = shape + '.f[%d]' % n;
		cmds.select(str(face_str))
		face_vertex = cmds.polyInfo(faceToVertex=True)
		# The regular expression (re) will convert this,
		# [u'FACE      0:      0      1      3      2 \n']
		# into this,
		# ['0', '0', '1', '3', '2']
		vertex_indexes = re.findall(r'\d+', str(face_vertex))
		vertex_indexes.pop(0)
		# After the pop() we now have this,
		# ['0', '1', '3', '2']
		
		face = []
		count = 0
		first_vert = []
		for index in vertex_indexes:
			vert_str = shape + '.vtx[%d]' % int(index);
			vert_pos = cmds.pointPosition(vert_str, local=useLocalSpace) 
			face.append(vert_pos)
			if count == 0:
				first_vert = vert_pos
			count += 1
		face.append(first_vert)
		faces.append(face)
	return faces
##=====================================
def get_edge_loops(tnode, useLocalSpace=True):
	return get_faces_verts(tnode, useLocalSpace=True)
	
##=====================================
# Thanks to Joseph Chen for supplying the code for the
# next function.
def get_faces_midpoint(tnode, useLocalSpace=True):
	faces = []
	shape = cmds.listRelatives(tnode, shapes=True)[0]
	num_faces = cmds.polyEvaluate(tnode, face=True)
	for n in range(num_faces):
		face_str = shape + '.f[%d]' % n;
		cmds.select(face_str)
		face_vertex = cmds.polyInfo(faceToVertex=True)
		face_vertex = re.findall(r'\d+', str(face_vertex))
		face_vertex.pop(0)
		verts_average = [0,0,0]
		for vert in face_vertex:
			vert_str = shape + '.vtx[%d]' % int(vert);
			vert_pos = cmds.pointPosition(vert_str, local=useLocalSpace) 
			verts_average[0]+=vert_pos[0]
			verts_average[1]+=vert_pos[1]
			verts_average[2]+=vert_pos[2]
		verts_average[0]= verts_average[0]/len(face_vertex)
		verts_average[1]= verts_average[1]/len(face_vertex)
		verts_average[2]= verts_average[2]/len(face_vertex)
		faces.append(verts_average)
	return faces
##=====================================
# Thanks to Joseph Chen for supplying the code for the
# next function.
def get_facesNormal(tnode, useLocalSpace=True):
	facesRotation = []
	shape = cmds.listRelatives(tnode, shapes=True)[0]
	num_faces = cmds.polyEvaluate(tnode, face=True)
	for n in range(num_faces):
		face_str = shape + '.f[%d]' % n;
		cmds.select(face_str)
		face_normal = cmds.polyInfo(faceNormals=True)
		face_normal = re.findall(r'[-+]?\d*\.\d+|\d+', str(face_normal))
		face_normal.pop(0)
		face_angle = cmds.angleBetween( euler=True, v1=(0.0, 0.1, 0.0), v2=face_normal)
		facesRotation.append(face_angle)
	return facesRotation
	
##=====================================
def distort(tnode, amount):
	selected_verts = cmds.filterExpand(selectionMask=31)
	random.seed(1)
	if selected_verts != None and len(selected_verts) > 0:
		for vertex in selected_verts:
			x = random.uniform(-amount/2, amount/2)
			y = random.uniform(-amount/2, amount/2)
			z = random.uniform(-amount/2, amount/2)
			cmds.polyMoveVertex(vertex, localTranslate=[x,y,z])
		cmds.select(selected_verts)
		return
		
	# No vertices selected - distort all the vertices
	shape_name = cmds.listRelatives(tnode, shapes=True)[0]
	num_verts = cmds.polyEvaluate(tnode, vertex=True)
	vertices = []
	for n in range(num_verts):
		vertex = shape_name + '.vtx[%d]' % n
		x = random.uniform(-amount/2, amount/2)
		y = random.uniform(-amount/2, amount/2)
		z = random.uniform(-amount/2, amount/2)
		cmds.polyMoveVertex(vertex, localTranslate=[x,y,z])
	cmds.select(clear=True)	

#========================================================
# Requires at least two objects. The first object specified by
# the name of the 'target_mesh' node MUST be polygon mesh. The 
# second parameter is a list of the names of the transform nodes
# of the shapes that will be used to cover the mesh. The only
# time the target_mesh needs to selected is when some of it's
# vertices have been selected. 
def cover_mesh(target_mesh, list_of_shapes, alignToNormals=True):
	vertices = get_verts(target_mesh, False)
	if alignToNormals == True:
		normals = get_normals(target_mesh)
	inst_names = []
	random.seed(1)
	for n in range(len(vertices)):
		vert_x,vert_y,vert_z = vertices[n]
		# randomly pick an instance from the list (insts)
		item = random.choice(list_of_shapes)
		inst = cmds.instance(item)[0]
		
		if alignToNormals == True:
			normal = normals[n]
			x_rot,z_rot = aim_y(normal)
			cmds.rotate(x_rot, 0, z_rot, inst)		
		cmds.move(vert_x,vert_y,vert_z, inst)
		inst_names.append(inst)
	return cmds.group(inst_names)
#========================================================
# Creates a collection of (straight) curves with CV's that match 
# the vertex positions of a polygon mesh. In effect the function
# creates a wireframe version of the mesh.
def make_wireframe(tnode, curve_width, shading_group_name=''):
	faces = get_edge_loops(tnode)
	names = []
	for face in faces:
		curve_name = cmds.curve(degree=1, p=face)
		# Add the renderman attributes so that the curve can
		# be rendered...
		cmds.setAttr(curve_name + ".rman_renderCurve", 1)
		cmds.setAttr(curve_name + ".rman_curveBaseWidth", curve_width)
		cmds.setAttr(curve_name + ".rman_curveTipWidth", curve_width)
		# Check if an SG has been assigned...
		if len(shading_group_name) > 0:
			cmds.connectAttr(shading_group_name + ".message", curve_name + ".rman__torattr___customShadingGroup")
		names.append(curve_name)
	cmds.group(names)
